/* This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/. */

/* global globalThis */

this.EXPORTED_SYMBOLS = ["mozCNUtils"];

ChromeUtils.defineModuleGetter(this, "XPCOMUtils",
  "resource://gre/modules/XPCOMUtils.jsm");
XPCOMUtils.defineLazyModuleGetters(this, {
  AddonManager: "resource://gre/modules/AddonManager.jsm",
  BackgroundPageThumbs: "resource://gre/modules/BackgroundPageThumbs.jsm",
  CustomizableUI: "resource:///modules/CustomizableUI.jsm",
  NewTabUtils: "resource://gre/modules/NewTabUtils.jsm",
  OS: "resource://gre/modules/osfile.jsm",
  PageThumbs: "resource://gre/modules/PageThumbs.jsm",
  PlacesUtils: "resource://gre/modules/PlacesUtils.jsm",
  WebChannel: "resource://gre/modules/WebChannel.jsm",
});

// Since Fx 104, see https://bugzil.la/1667455,1780695
const Services =
  globalThis.Services ||
  ChromeUtils.import("resource://gre/modules/Services.jsm").Services;

XPCOMUtils.defineLazyGetter(this, "gMM", () => {
  return Cc["@mozilla.org/globalmessagemanager;1"].
    getService(Ci.nsIMessageListenerManager || Ci.nsISupports);
});


XPCOMUtils.defineLazyModuleGetters(this, {
  delayedSuggestBaidu: "resource://ntab/mozCNUtils.jsm",
  Frequent: "resource://ntab/mozCNUtils.jsm",
  getPref: "resource://ntab/mozCNUtils.jsm",
  Homepage: "resource://ntab/mozCNUtils.jsm",
  NTabDB: "resource://ntab/NTabDB.jsm",
  NTabWindow: "resource://ntab/NTabWindow.jsm",
  Session: "resource://ntab/mozCNUtils.jsm",
  Tracking: "resource://ntab/Tracking.jsm",
});

this.strings = {
  _ctx: null,

  init(context) {
    this._ctx = context;
  },

  uninit() {
    delete this._ctx;
  },

  _(name, subs) {
    if (!this._ctx) {
      return "";
    }

    let cloneScope = this._ctx.cloneScope;
    return this._ctx.extension.localizeMessage(name, subs, {cloneScope});
  },
};

this.morePermissionPromptHack = {
  extensionId: "",
  prefKey: "extensions.chinaEditionHomepage.pendingNextVersion",
  topic: "webextension-update-permissions",

  get nextVersion() {
    return Services.prefs.getCharPref(this.prefKey, "0");
  },
  set nextVersion(version) {
    if (version) {
      Services.prefs.setCharPref(this.prefKey, version);
    } else {
      Services.prefs.clearUserPref(this.prefKey);
    }
  },

  async init({extension}) {
    this.extensionId = extension.id;
    Services.obs.addObserver(this, this.topic);

    if (this.nextVersion &&
        Services.vc.compare(extension.version, this.nextVersion) >= 0) {
      this.nextVersion = null;
      return;
    }

    let addon = await AddonManager.getAddonByID(this.extensionId);
    addon.findUpdates({
      onUpdateAvailable(addon, install) {
        if (addon.permissions & AddonManager.PERM_CAN_UPGRADE &&
            AddonManager.shouldAutoUpdate(addon)) {
          // Trigger the installation w/o the permission prompt
          install.install();
        }
      },
    }, AddonManager.UPDATE_WHEN_PERIODIC_UPDATE);
  },

  observe(subject, topic, data) {
    if (topic !== this.topic) {
      return;
    }

    let { addon, type } = subject.wrappedJSObject;
    if (addon.id !== this.extensionId || type !== "update") {
      return;
    }

    this.nextVersion = addon.version;
  },

  uninit(isAppShutdown) {
    if (isAppShutdown) {
      return;
    }

    Services.obs.removeObserver(this, this.topic);
  },
};

this.newtabMigration = {
  dialPref: "browser.newtabpage.pinned",
  extId: "",
  otherExtId: "china-newtab@mozillaonline.com",
  otherExtPref: "extensions.chinaNewtab.prefVersion",
  otherExtUrl: "https://download-ssl.firefox.com.cn/chinaedition/addons/china-newtab/china-newtab-latest.xpi?cachebust=20200616",

  STATUS_BLOCKED: 0,
  STATUS_COMPAT: 1,
  STATUS_READY: 2,
  STATUS_PENDING: 3,
  STATUS_DONE: 4,

  statusPref: "extensions.cehomepage.splitNewtab",

  thumbnailBase: "https://offlintab.firefoxchina.cn",

  get isInstalled() {
    return Services.prefs.getIntPref(this.otherExtPref, 0) > 0;
  },

  get searchTN() {
    let searchTN = "error";
    try {
      let engine = Services.search.getEngineByName("\u767e\u5ea6");
      let newtabUrl = engine.getSubmission("TEST", null, "newtab").uri.spec;
      searchTN = (new URL(newtabUrl)).searchParams.get("tn") || "notset";
    } catch (ex) {
      console.error(ex);
      return searchTN;
    }
    delete this.searchTN;
    return this.searchTN = searchTN;
  },

  get status() {
    return Services.prefs.getIntPref(this.statusPref, this.STATUS_BLOCKED);
  },

  set status(status) {
    let oldStatus = this.status;
    let newStatus = Math.max(oldStatus, status);
    Services.prefs.setIntPref(this.statusPref, newStatus);
    try {
      if (newStatus > oldStatus) {
        Tracking.track({
          type: "split-newtab",
          action: "advance",
          sid: `${newStatus}`,
          fid: `${oldStatus}`,
        });
      }
    } catch (ex) {
      console.error(ex);
    }
  },

  async init({ extension }) {
    this.extId = extension.id;

    if (this.isInstalled) {
      this.status = this.STATUS_DONE;
    }

    if (this.status >= this.STATUS_DONE) {
      return;
    }

    if (Services.vc.compare(Services.appinfo.version, "68.0") < 0) {
      return;
    }
    // Don't override existing customizations, at least initially.
    if (Services.prefs.prefHasUserValue(this.dialPref)) {
      try {
        let dials = JSON.parse(Services.prefs.getStringPref(this.dialPref));
        let thumbnailDir = "https://offlintab.firefoxchina.cn/data/thumbnails/";
        if (!dials.every(dial => {
          if (dial === null) {
            return true;
          }

          if (dial.searchTopSite === true) {
            return true;
          }

          return (dial.customScreenshotURL || "").startsWith(thumbnailDir);
        })) {
          return;
        }
      } catch (ex) {
        console.error(ex);
        return;
      }
    }
    // Skip non-official search codes, for now.
    await Services.search.init();
    if (!/monline_[347]_dg/.test(this.searchTN)) {
      return;
    }

    this.status = this.STATUS_COMPAT;
    Services.prefs.addObserver(this.otherExtPref, this);
  },

  uninit(isAppShutdown) {
    if (isAppShutdown) {
      return;
    }

    try {
      Services.prefs.removeObserver(this.otherExtPref, this);
    } catch (ex) {
      console.error(ex);
    }
  },

  async installExt() {
    if (this.status !== this.STATUS_READY) {
      return;
    }

    // Suppress the prompt for reverting to "about:newtab"
    let prefKey = `extensions.installedDistroAddon.${this.otherExtId}`;
    Services.prefs.setBoolPref(prefKey, true);

    let install = await AddonManager.getInstallForURL(this.otherExtUrl, {
      telemetryInfo: { source: this.extId },
    });

    // Mostly from /browser/components/enterprisepolicies/Policies.jsm
    if (install.addon && install.addon.appDisabled) {
      install.cancel();
      return;
    }

    let listener = {
      onDownloadEnded: install => {
        if (install.addon.id !== this.otherExtId) {
          install.removeListener(listener);
          install.cancel();
        }
        if (install.addon && install.addon.appDisabled) {
          install.removeListener(listener);
          install.cancel();
        }
        if (install.addon &&
            Services.vc.compare(install.addon.version, "4.77.4") < 0) {
          install.removeListener(listener);
          install.cancel();
        }
      },
      onDownloadFailed: () => {
        install.removeListener(listener);
      },
      onInstallEnded: () => {
        install.removeListener(listener);
        this.status = this.STATUS_PENDING;
      },
      onInstallFailed: () => {
        install.removeListener(listener);
      },
    };
    install.addListener(listener);
    install.install();
  },

  async migrate(data) {
    try {
      Tracking.track({
        type: "split-newtab",
        action: "migrate",
        sid: `${this.status}`, // Number 0 is a falsy value
        fid: `${Services.appinfo.version}-${this.searchTN}`,
      });
    } catch (ex) {
      console.error(ex);
    }

    if (this.status <= this.STATUS_BLOCKED ||
        this.status >= this.STATUS_DONE) {
      return;
    }

    let {
      dialContent,
      "dial.hideSearch": hideSearch,
      "search.engine": searchEngine,
      view,
    } = JSON.parse(data.state);

    this.migrateDials(dialContent);
    this.migrateSearch(hideSearch, searchEngine);
    this.migrateView(view);

    this.status = this.STATUS_READY;
    await this.installExt();
  },

  migrateDials(dials) {
    let pinned = [];
    let keys = Object.keys(dials);
    for (let key of keys) {
      let dial = dials[key];

      let item = {
        label: dial.title,
        url: dial.url,
      };
      let defaultPosition = parseInt(dial.defaultposition, 10) - 1;
      if (dial.thumbnail && !isNaN(defaultPosition)) {
        item.customScreenshotURL = `${this.thumbnailBase}${dial.thumbnail}?mococn_dp=${defaultPosition}`;
      }
      pinned[parseInt(key, 10) - 1] = item;
    }
    Services.prefs.setStringPref(this.dialPref, JSON.stringify(pinned));

    // Do we need to clear the `TopSitesFeed` cache here ?
    NewTabUtils.pinnedLinks.resetCache();
  },

  // Not sure we want these two ?
  migrateSearch(hideSearch, searchEngine) {},
  migrateView(view) {},

  observe(subject, topic, data) {
    switch (topic) {
      case "nsPref:changed":
        let otherExtPref = newtabMigration.otherExtPref;
        if (data !== otherExtPref) {
          break;
        }

        if (!newtabMigration.isInstalled) {
          break;
        }
        newtabMigration.status = newtabMigration.STATUS_DONE;
        Services.prefs.removeObserver(otherExtPref, newtabMigration);
        break;
      default:
        break;
    }
  },
};

this.mozCNUtils = {
  QueryInterface: ChromeUtils.generateQI([Ci.nsIObserver,
                                          Ci.nsIMessageListener]),

  // nsIObserver
  observe(aSubject, aTopic, aData) {
    switch (aTopic) {
      case "http-on-examine-response":
      case "http-on-examine-cached-response":
      case "http-on-examine-merged-response":
        this.trackHTTPStatus(aSubject, aTopic);
        break;
    }
  },

  frameScripts: [
    "resource://ntab/ntabContent.js",
  ],
  initFrameScripts() {
    this.frameScripts.forEach(frameScript => {
      gMM.loadFrameScript(frameScript, true);
    });
  },
  uninitFrameScripts() {
    this.frameScripts.forEach(frameScript => {
      gMM.removeDelayedFrameScript(frameScript);
    });
  },

  trackHTTPStatus(aSubject, aTopic) {
    let channel = aSubject;
    channel.QueryInterface(Ci.nsIHttpChannel);

    if (![
      NTabDB.prePath,
      "http://i.g-fox.cn",
      "https://home.firefoxchina.cn",
    ].includes(channel.URI.prePath) ||
        !(channel.loadFlags & Ci.nsIChannel.LOAD_DOCUMENT_URI)) {
      return;
    }

    if (![200, 302, 304].includes(channel.responseStatus)) {
      Tracking.track({
        type: "http-status",
        sid: channel.responseStatus,
        action: aTopic,
        href: channel.URI.spec,
        altBase: "http://robust.g-fox.cn/ntab.gif",
      });
    }
  },

  // nsIMessageListener
  receiveMessage(aMessage) {
    if (!this.MESSAGES.includes(aMessage.name)) {
      return;
    }

    switch (aMessage.name) {
      case "mozCNUtils:NTabSync":
        newtabMigration.migrate(aMessage.data);
        break;
      case "mozCNUtils:Tracking":
        Tracking.track(aMessage.data);
        break;
      case "mozCNUtils:WebChannel":
        switch ((aMessage.data && aMessage.data.type)) {
          case "isFxDefaultBrowser":
            // a response should be sent even shellService not available
            aMessage.target.messageManager.sendAsyncMessage(aMessage.name, {
              type: aMessage.data.type,
              data: (this.shellService &&
                     this.shellService.isDefaultBrowser(false, false)),
            });
            break;
          case "setFxAsDefaultBrowser":
            if (!this.shellService) {
              break;
            }
            this.shellService.setDefaultBrowser(false, false);
            break;
        }
        break;
    }
  },

  get shellService() {
    let shellService;
    try {
      shellService = Cc["@mozilla.org/browser/shell-service;1"].
        getService(Ci.nsIShellService);
    } catch (e) {
      // shellService is not universally available, see https://bugzil.la/297841
    }
    delete this.shellService;
    return this.shellService = shellService;
  },

  MESSAGES: [
    "mozCNUtils:NTabSync",
    "mozCNUtils:Tracking",
    "mozCNUtils:WebChannel",
  ],
  initMessageListener() {
    for (let msg of this.MESSAGES) {
      gMM.addMessageListener(msg, this);
    }
  },
  uninitMessageListener() {
    for (let msg of this.MESSAGES) {
      gMM.removeMessageListener(msg, this);
    }
  },

  // TabsProgressListener variant of nsIWebProgressListener
  onStateChange(aBrowser, aWebProgress, aRequest, aStateFlags, aStatus) {
    if (aWebProgress.isTopLevel &&
        (aStateFlags & Ci.nsIWebProgressListener.STATE_IS_WINDOW)) {
      let isStart = aStateFlags & Ci.nsIWebProgressListener.STATE_START;
      let isStop = aStateFlags & Ci.nsIWebProgressListener.STATE_STOP;
      if (!isStart && !isStop) {
        return;
      }

      if (delayedSuggestBaidu.isGoogleSearch(aRequest.URI)) {
        if (isStart) {
          delayedSuggestBaidu.attach(aBrowser, aRequest.URI);
        }
        if (isStop) {
          delayedSuggestBaidu.remove(aBrowser, aStatus);
        }
      }
    }
  },

  onLocationChange(aBrowser, b, aRequest, aLocation, aFlags) {
    if (aFlags & Ci.nsIWebProgressListener.LOCATION_CHANGE_ERROR_PAGE) {
      // before we can fix the OfflineCacheInstaller ?
      if (aLocation.equals(NTabDB.uri)) {
        aRequest.cancel(Cr.NS_BINDING_ABORTED);
        aBrowser.webNavigation.loadURI("about:blank", null, null, null, null);
      }
    }
  },

  initDefaultPrefs() {
    let defBranch = Services.prefs.getDefaultBranch("");

    defBranch.setBoolPref("moa.ntab.openInNewTab", true);
    defBranch.setBoolPref("services.sync.engine.mozcn.ntab", false);
  },

  initWindowListener() {
    for (let win of CustomizableUI.windows) {
      this.onWindowOpened(win);
    }

    CustomizableUI.addListener(this);
  },

  uninitWindowListener() {
    CustomizableUI.removeListener(this);

    for (let win of CustomizableUI.windows) {
      this.onWindowClosed(win);
    }
  },

  onWindowOpened(win) {
    if (win.gBrowser) {
      win.gBrowser.addTabsProgressListener(this);
    } else if (win._gBrowser) {
      win._gBrowser.addTabsProgressListener(this);
    } else {
      win.console.error("Neither gBrowser or _gBrowser ?");
    }
    NTabWindow.onWindowOpened(win);
  },

  onWindowClosed(win) {
    win.gBrowser.removeTabsProgressListener(this);
    NTabWindow.onWindowClosed(win);
  },

  init(context) {
    morePermissionPromptHack.init(context);
    let isAppStartup = context.extension.startupReason === "APP_STARTUP";
    strings.init(context);

    Services.obs.addObserver(this, "http-on-examine-response");
    Services.obs.addObserver(this, "http-on-examine-cached-response");
    Services.obs.addObserver(this, "http-on-examine-merged-response");

    this.initDefaultPrefs();
    this.initFrameScripts();
    this.initMessageListener();

    delayedSuggestBaidu.init(strings);
    Homepage.init(isAppStartup);
    mozCNWebChannels.init();
    newtabMigration.init(context);
    NTabWindow.init(strings);

    // this needs to run after NTabWindow.init for strings
    this.initWindowListener();
  },

  uninit(isAppShutdown) {
    Services.obs.removeObserver(this, "http-on-examine-response");
    Services.obs.removeObserver(this, "http-on-examine-cached-response");
    Services.obs.removeObserver(this, "http-on-examine-merged-response");

    this.uninitFrameScripts();
    this.uninitMessageListener();
    this.uninitWindowListener();

    Homepage.uninit(isAppShutdown);
    mozCNWebChannels.uninit();
    morePermissionPromptHack.uninit(isAppShutdown);
    newtabMigration.uninit(isAppShutdown);
    NTabWindow.uninit();

    for (let jsModule of [
      "mozCNUtils",
      "NTabDB",
      "NTabWindow",
      "Tracking",
    ]) {
      try {
        Cu.unload(`resource://ntab/${jsModule}.jsm`);
      } catch (ex) {
        console.error(ex);
      }
    }
  },
};

this.mozCNWebChannel = function(aChannelID, aURI, aListener) {
  this.channel = new WebChannel(aChannelID, aURI);
  this.channel.listen((this[aListener] || this.baseListener).bind(this));
};
this.mozCNWebChannel.prototype = {
  baseListener(a, aMessage, aSender) {
    switch (aMessage.key) {
      case "frequent.query":
        Frequent.query(aEntries => {
          this.channel.send({
            id: aMessage.id,
            key: aMessage.key,
            data: aEntries,
          }, aSender);
        }, aMessage.parameters.limit);
        break;
      case "frequent.remove":
        Frequent.remove(() => {
          this.channel.send({
            id: aMessage.id,
            key: aMessage.key,
          }, aSender);
        }, [aMessage.parameters.url]);
        break;
      case "frequent.tophosts":
        Frequent.topHosts(aEntries => {
          this.channel.send({
            id: aMessage.id,
            key: aMessage.key,
            data: aEntries,
          }, aSender);
        }, aMessage.parameters.hosts);
        break;
      case "last.query":
        Session.query(aEntries => {
          this.channel.send({
            id: aMessage.id,
            key: aMessage.key,
            data: aEntries,
          }, aSender);
        }, aMessage.parameters.limit);
        break;
      case "last.remove":
        Session.remove(() => {
          this.channel.send({
            id: aMessage.id,
            key: aMessage.key,
          }, aSender);
        }, [aMessage.parameters.url]);
        break;
      case "startup.channelid":
        this.channel.send({
          id: aMessage.id,
          key: aMessage.key,
          data: getPref("app.chinaedition.channel", "www.firefox.com.cn"),
        }, aSender);
        break;
    }
  },
  offlintabListener(a, aMessage, aSender) {
    this.baseListener.apply(this, arguments);

    let self = this;
    switch (aMessage.key) {
      case "bookmark.query":
        let links = [];
        PlacesUtils.promiseDBConnection().then(db => {
          return db.executeCached(`
SELECT
  b.title as title,
  p.url as url
FROM
  moz_bookmarks b,
  moz_places p
WHERE
  b.type = 1 AND
  b.fk = p.id AND
  p.hidden = 0
`,
            null,
            row => {
              links.push({
                title: row.getResultByName("title"),
                url: row.getResultByName("url"),
              });
            }
          );
        }).then(() => {
          self.channel.send({
            id: aMessage.id,
            key: aMessage.key,
            data: links,
          }, aSender);
        }, ex => {
          console.error(ex);
          self.channel.send({
            id: aMessage.id,
            key: aMessage.key,
            data: [],
          }, aSender);
        });
        break;
      case "thumbs.getThumbnail":
        /**
         * use capture instead of captureIfMissing to force generate the
         * better looking version.
         */
        let url = aMessage.parameters.url;
        BackgroundPageThumbs.capture(url, {
          onDone() {
            let path = PageThumbs.getThumbnailPath(url);
            OS.File.read(path).then(aData => {
              let blob = new Blob([aData], {
                type: PageThumbs.contentType,
              });
              self.channel.send({
                id: aMessage.id,
                key: aMessage.key,
                data: {
                  url,
                  blob,
                },
              }, aSender);
            }, aError => {
              self.channel.send({
                id: aMessage.id,
                key: aMessage.key,
                data: {
                  url,
                },
              }, aSender);
            });
          },
        });
        break;
      case "variant.channel":
        this.channel.send({
          id: aMessage.id,
          key: aMessage.key,
          data: "master-ii",
        }, aSender);
        break;
    }
  },
};

this.mozCNWebChannels = {
  channelID: "moz_cn_channel_v2",
  contentURL: "resource://ntab/mozCNWebChannelContent.js",
  specs: {
    "https://home.firefoxchina.cn/": "",
    "https://i.firefoxchina.cn/": "",
    "http://newtab.firefoxchina.cn/": "",
    "https://newtab.firefoxchina.cn/": "",
    "http://offlintab.firefoxchina.cn/": "offlintabListener",
    "https://offlintab.firefoxchina.cn/": "offlintabListener",
  },
  webChannels: [],
  init() {
    Object.keys(this.specs).forEach(aSpec => {
      let uri = Services.io.newURI(aSpec);
      let webChan = new mozCNWebChannel(this.channelID, uri, this.specs[aSpec]);
      this.webChannels.push(webChan);
    });
    gMM.loadFrameScript(this.contentURL, true);
  },
  uninit() {
    while (this.webChannels.length) {
      let webChan = this.webChannels.shift();
      webChan.channel.stopListening();
    }
    gMM.removeDelayedFrameScript(this.contentURL);
  },
};
